#include "terminal_api_driver.h"

#include <algorithm>

#include "terminal_api/ControlProtoCInterface.h"
#include <sstream>
#include <cmath>
#include <iomanip>

/*
timer_->timeout().connect([=]{
    double txpwd = 0;
    CP_GetGain(sid, "TXPWD", &txpwd);
    transmitt_active->set_end(!txpwd);
    CP_GetLevelDemod(sid, "modcod_tx", &txpwd);
    modcod_tx->setText(std::to_string(static_cast<uint32_t>(txpwd) >> 2));
    uint8_t type_pack = static_cast<uint8_t>(txpwd);
    type_pack = type_pack & 0b00000010;
    if(type_pack)
        type_pack_lbl->setText("short");
    else
        type_pack_lbl->setText("normal");

    uint8_t is_pilots = static_cast<uint8_t>(txpwd);
    is_pilots = is_pilots & 0b00000001;
    if(is_pilots)
        pilot_lbl->setText("pilots");
    else
        pilot_lbl->setText("no pilots");

    CP_GetLevelDemod(sid, "snr_acm", &txpwd);
    std::stringstream buf_double;
    buf_double << std::fixed <<
    std::setprecision(2) << txpwd;
    snr_acm->setText(buf_double.str());
});

 */

api_driver::ApiDriver::ApiDriver() {
    CP_Login("admin", "pass", &sid, &access);
    CP_GetDmaDebug(sid, "status_init", &deviceInitState);
}

static const char* boolAsStr(bool value) {
    return value ? "true" : "false";
}

static std::string buildEscapedString(const std::string& source) {
    std::string str(source);
    size_t start_pos = 0;
    while((start_pos = str.find('\"', start_pos)) != std::string::npos) {
        str.replace(start_pos, 1, "\\\"");
        start_pos += 2;
    }
    return "\"" + str + "\"";
}

static bool DriverCP_GetCinC(TSID sid) {
    modulator_settings s{};
    CP_GetModulatorSettings(sid, s);
    return s.is_cinc;
}

std::string api_driver::ApiDriver::loadTerminalState() {
    std::stringstream result;
    result << "{\"initState\":" << buildEscapedString(this->deviceInitState);

    modulator_state modulator{};
    CP_GetModulatorState(sid, modulator);

    demodulator_state demodulator{};
    CP_GetDemodulatorState(sid, demodulator);

    device_state device{};
    CP_GetDeviceState(sid, device);

    const bool isCinC = DriverCP_GetCinC(sid);

    result << ",\"isCinC\":" << boolAsStr(isCinC);

    // формируем структуру для TX
    result << ",\"tx.state\":" << boolAsStr(modulator.is_tx_on);
    result << ",\"tx.modcod\":" << modulator.modcod;
    result << ",\"tx.snr\":" << std::fixed << std::setprecision(2) << modulator.snr_remote;

    if (modulator.is_short) {
        result << R"(,"tx.frameSize":"short")";
    } else {
        result << R"(,"tx.frameSize":"normal")";
    }

    if (modulator.is_pilots) {
        result << R"(,"tx.pilots":"pilots")";
    } else {
        result << R"(,"tx.pilots":"no pilots")";
    }

    result << ",\"tx.speedOnTxKbit\":" << std::fixed << std::setprecision(2) << static_cast<double>(modulator.speed_in_bytes_tx) / 128.0;
    result << ",\"tx.speedOnIifKbit\":" << std::fixed << std::setprecision(2) << (static_cast<double>(modulator.speed_in_bytes_tx_iface) / 128.0);

    // формируем структуру для RX
    result << ",\"rx.state\":" << boolAsStr(demodulator.locks.sym_sync_lock && demodulator.locks.freq_lock && demodulator.locks.afc_lock && demodulator.locks.pkt_sync);
    result << ",\"rx.sym_sync_lock\":" << boolAsStr(demodulator.locks.sym_sync_lock);
    result << ",\"rx.freq_search_lock\":" << boolAsStr(demodulator.locks.freq_lock);
    result << ",\"rx.afc_lock\":" << boolAsStr(demodulator.locks.afc_lock);
    result << ",\"rx.pkt_sync\":" << boolAsStr(demodulator.locks.pkt_sync);

    result << ",\"rx.snr\":" << std::fixed << std::setprecision(2) << demodulator.snr;
    result << ",\"rx.rssi\":" << std::fixed << std::setprecision(2) << demodulator.rssi;
    result << ",\"rx.modcod\":" << demodulator.modcod;

    if (demodulator.is_short) {
        result << R"(,"rx.frameSize":"short")";
    } else {
        result << R"(,"rx.frameSize":"normal")";
    }

    if (demodulator.is_pilots) {
        result << R"(,"rx.pilots":"pilots")";
    } else {
        result << R"(,"rx.pilots":"no pilots")";
    }

    result << ",\"rx.symError\":" << std::fixed << std::setprecision(2) << demodulator.sym_err;
    result << ",\"rx.freqErr\":" << std::fixed << std::setprecision(2) << demodulator.crs_freq_err;
    result << ",\"rx.freqErrAcc\":" << std::fixed << std::setprecision(2) << demodulator.fine_freq_err;
    result << ",\"rx.inputSignalLevel\":" << std::fixed << std::setprecision(2) << demodulator.if_overload;
    result << ",\"rx.pllError\":" << std::fixed << std::setprecision(2) << demodulator.afc_err;
    result << ",\"rx.speedOnRxKbit\":" << std::fixed << std::setprecision(2) << (static_cast<double>(demodulator.speed_in_bytes_rx) / 128.0);
    result << ",\"rx.speedOnIifKbit\":" << std::fixed << std::setprecision(2) << (static_cast<double>(demodulator.speed_in_bytes_rx_iface) / 128.0);
    result << ",\"rx.packetsOk\":" << demodulator.packet_ok_cnt;
    result << ",\"rx.packetsBad\":" << demodulator.packet_bad_cnt;
    result << ",\"rx.packetsDummy\":" << demodulator.dummy_cnt;

    // формируем структуру для CinC
    if (isCinC) {
        CinC_state state_cinc{};
        CP_GetCinCState(sid,state_cinc);

        if (modulator.is_tx_on) {
            if (state_cinc.carrier_lock) {
                result << R"(,"cinc.correlator":true)";
            } else {
                result << R"(,"cinc.correlator":false)";
            }
        } else {
            result << R"(,"cinc.correlator":null)";
        }

        result << ",\"cinc.occ\":" << std::fixed << std::setprecision(3) << state_cinc.ratio_signal_signal;
        result << ",\"cinc.correlatorFails\":" << state_cinc.cnt_bad_lock;
        result << ",\"cinc.freqErr\":" << state_cinc.freq_error_offset;
        result << ",\"cinc.freqErrAcc\":" << state_cinc.freq_fine_estimate;
        result << ",\"cinc.channelDelay\":" << state_cinc.delay_dpdi;
    } else {
        result << R"(,"cinc.correlator":null)";
    }

    // структура температур девайса
    result << ",\"device.adrv\":" << std::fixed << std::setprecision(3) << device.adrv_temp;
    result << ",\"device.fpga\":" << std::fixed << std::setprecision(3) << device.pl_temp;
    result << ",\"device.zync\":" << std::fixed << std::setprecision(3) << device.zynq_temp;

    result << "}";

    return result.str();
}


void api_driver::ApiDriver::resetPacketStatistics() const {
    std::string tmp;
    CP_GetDmaDebug(sid, "reset_cnt_rx", &tmp);
}

std::string api_driver::ApiDriver::loadSettings() {
    // modulator_settings modSettings{};
    // CP_GetModulatorSettings(sid, modSettings);
    //
    // modSettings.baudrate = 2000000;
    // modSettings.central_freq_in_kGz = 1340000.24;
    // modSettings.rollof = 0.25;
    // modSettings.tx_is_on = true;
    // modSettings.is_test_data = false;
    // modSettings.is_save_current_state = true;
    // modSettings.is_cinc = false;
    // modSettings.is_carrier = true;
    // CP_SetModulatorSettings(sid, modSettings);
    // demodulator_settings demodulator_sett;
    // demodulator_sett.baudrate = 1340000;
    // demodulator_sett.central_freq_in_kGz = 2400000.34;
    // demodulator_sett.is_rvt_iq = true;
    // demodulator_sett.is_aru_on = true;
    // demodulator_sett.gain = 13;
    // demodulator_sett.rollof = 0.15;
    // if(CP_SetDemodulatorSettings(sid, demodulator_sett)== OK)
    // {
    //     std::cout << "OK SET DEMODULATOR SETTINGS" <<  std::endl;
    // }
    // demodulator_settings demodulator_settings_;
    // if(CP_GetDemodulatorSettings(sid,demodulator_settings_) == OK)
    // {
    //     std::cout << "OK GET DEMODULATOR SETTINGS" << std::endl;
    //     std::cout << demodulator_settings_.baudrate << std::endl;
    //     std::cout << demodulator_settings_.gain<< std::endl;
    //     std::cout << demodulator_settings_.rollof << std::endl;
    //     std::cout << demodulator_settings_.is_aru_on << std::endl;
    //     std::cout << demodulator_settings_.is_rvt_iq << std::endl;
    //     std::cout << demodulator_settings_.central_freq_in_kGz << std::endl;
    // }
    // modulator_settings modulator_settings_;
    // if(CP_GetModulatorSettings(sid,modulator_settings_)== OK)
    // {
    //     std::cout << "OK GET MODULATOR SETTINGS" << std::endl;
    //     std::cout << modulator_settings_.baudrate << std::endl;
    //     std::cout << modulator_settings_.attenuation << std::endl;
    //     std::cout << modulator_settings_.central_freq_in_kGz << std::endl;
    //     std::cout << modulator_settings_.is_carrier << std::endl;
    //     std::cout << modulator_settings_.is_cinc << std::endl;
    //     std::cout << modulator_settings_.rollof << std::endl;
    // }

    return "{}";
}

api_driver::ApiDriver::~ApiDriver() = default;
